#!/usr/bin/env python
#Classe de configuration

from UserDict import UserDict
from types    import *
import pickle,os,ConfigParser
import logging

logger = logging.getLogger('wabaConf  ')

class dataConf(UserDict):
   def __init__(self,dict=None):
      UserDict.__init__(self,dict)

class wabaConf:
   """ COnfiguration class
   to be used like a dictionary
   parameters can be read from wabaroK config file, or from amaroK config filename
   if read from wabaroK conf, they are saved in it. else they are not
   eg: dbengine is the amarok database engine (MySQL or SqlLite), normaly  read
       from ~/.kde/share/config/amarokrc:DatabaseEngine If you want to modify it, add a dbengine=1 in wabarok.cfg
       then it will be read from it instead of amarokrc"""
   # params from conf file
   #        key:          ("Displayed name",    "kdialog to enter value",      ValueType,  DefValue, saved?),
   params={'port':        ["Listening port",    "%s --inputbox \"%s\" \"%s\" ",IntType,    1234,     1],
           'theme':       ["Theme",             "%s --inputbox \"%s\" \"%s\" ",StringType, "default",1],
           'imagewidth':  ["Cover Width",       "%s --inputbox \"%s\" \"%s\" ",StringType, "200",    1],
           'imageheight': ["Cover Height",      "%s --inputbox \"%s\" \"%s\" ",StringType, "200",    1],
           'column':      ["Number of Columns" ,"%s --inputbox \"%s\" \"%s\" ",IntType,     3,       1],
           'logging':     ["Log level",         "%s --radiolist \"%s\" 0 Debug 0 1 Info 0 2 Warning 1 3 Errors 0 4 Critical 0 %s",
                                                                               IntType,     3,       1],
           'bitrate':     ["Streaming bitrate (kbps)", "%s --radiolist \"%s\" 8 8 0 16 16 0 32 32 1 48 48 0 56 56 0 64 64 0 96 96 0 %s",
                                                                               IntType,     3,       1],
         # not to be saved, normaly read from amarokrc, or hidden parameter
           'dbengine':    ["DatabaseEngine",    "%s --inputbox \"%s\" \"%s\" ",IntType,     0,       0],
           'collfolders': ["Collection Folders","%s --inputbox \"%s\" \"%s\" ",ListType,   [],       0],
           'covers'     : ["Collection Folders","%s --inputbox \"%s\" \"%s\" ",StringType, '',       0]
          }
   # logging: logging values (mapped from logging module / 10)
   #  CRITICAL 5 : everything has to be stopped
   #  ERROR    4 : big errors
   #  WARNING  3 : something seems to be wrong, or need to be known
   #  INFO     2 : usefull informations for dev and tests
   #  DEBUG    1 : values inside loop, big dumps...
   #  NOTSET   0

   def __init__(self,localprefix,dataPath,filename="wabaroK.cfg"):
      logger.info("__init__(%s,%s,%s)",localprefix,dataPath,filename)
      self.data=dataConf()
      self.data['configfile']=filename
      self.data['dataPath'] = dataPath
      self.localprefix=localprefix
      self.kdialog = os.popen('which kdialog').readline().rstrip()
      if not os.access(self.kdialog,os.X_OK):
         logger.critical("Cannot find kdialog, you wont be able to change configuration using this GUI (%s)",self.kdialog)
      else:
         logger.info("using %s",self.kdialog)


   def readPickle(self):
      """read from config file
      ** outdated **"""
      logger.error('readPickle: outdated function, we should never go in')
      try:
         f=open(self.data['configfile'],"r")
         self.data=pickle.load(f)
         f.close()
      except Exception, detail:
         # cannot read conf file, default values
         logging.warning("cannot read wabaroK config file "+str(Exception)+" "+str(detail))
         self.data['port']=1234
         self.data['theme']='default'
         self.data['imageWidth']="200"
         self.data['imageHeight']="200"
         self.data['column']=3
         self.data['logging']=3

   def readText(self):
      """Read parameters from text config file"""
      logger.info("readText() in %s",self.data['configfile'])
      waConf=ConfigParser.SafeConfigParser()
      # Get saved parameters
      try:
         waConf.readfp(open(self.data['configfile'],"r"))
         for k in self.params.keys():
            try: # get param from section General
               if self.params[k][2] == IntType:
                  self.data[k]=waConf.getint("General",k)
               elif self.params[k][2] == ListType:
                  # split on , and remove trailing spaces
                  self.data[k]=[a.strip() for a in waConf.get("General",k).split(',')]
               else:
                  self.data[k]=waConf.get("General",k)
               # read from config file without exception, so set it to be saved
               self.params[k][4] = 1
            except Exception, detail:
               logger.info("readText: %s not set in wabaroK configuration file (%s)",k,str(detail))
               # Parameter not set in file,
               if self.params[k][4]:
                  # if tobesaved, set it to default
                  self.data[k]=self.params[k][3]
      except Exception, detail:
         # Cannot read config file, set all to defaults
         logger.warning("readText: cannot read config file %s (%s)",self.data['configfile'],str(detail))
         logger.warning("readText: set everything to defaults")
         for k in self.params.keys():
            if self.params[k][4]: # if to be saved
               self.data[k]=self.params[k][3] #set to default
      # get installed directory
      if not self.data.has_key('dbengine'):
         # not read in config file
         self.setAmarokDb(self.localprefix)
      if not self.data.has_key('collfolders'):
         self.setCollectionPath(self.localprefix)
      if not self.data.has_key('covers'):
         self.setAmazonCoverPath(self.localprefix)


   def read(self):
      #self.readPickle()
      self.readText()
   
   def setAmarokDb(self,localprefix):
      """get Db configured in amaroK
      ~/.kde/share/config/amarokrc
         [Collection]
         DatabaseEngine=?
      IN: localprefix ~/.kde/"""
      logger.info("setAmarokDb(%s)",localprefix)
      fn=localprefix+"/share/config/amarokrc"
      try:
         # 1st we try te get value from amarokrc
         amConf = ConfigParser.SafeConfigParser()
         amConf.readfp(open(fn))
         #if amConf.has_option("Collection","DatabaseEngine"):
         self.data["dbengine"]=amConf.getint("Collection","DatabaseEngine")
      except Exception,detail:
         # we cannot read amaroK config file or no DbEngine value
         logger.error("setAmarokDb: cannot read amaroK config file "+str(detail))
         if not self.data.has_key("dbengine"):
            # there is no value in wabaroK config
            logger.error("setAmarokDb: Nothing in wabaroK config file, setting Database engine SQLite (0)")
            self.data["dbengine"]=0
         else:
            logger.warning("setAmarokDb: Database engine read from wabaroK config file (%d)",self.data["DatabaseEngine"])
         
   def setCollectionPath(self,localprefix):
      """get Collection paths configured in amaroK
      ~/.kde/share/config/amarokrc
         [Collection]
         Collection Folders=/path/1,/path/2,...
      IN: localprefix ~/.kde/"""
      logger.info("setCollectionPath(%s)",localprefix)
      fn=localprefix+"/share/config/amarokrc"
      try:
         # 1st we try te get value from amarokrc
         amConf = ConfigParser.SafeConfigParser()
         amConf.readfp(open(fn))
         #if amConf.has_option("Collection","DatabaseEngine"):
         list=amConf.get("Collection","Collection Folders")
         self.data["collfolders"]=list.split(',')
      except Exception,detail:
         # we cannot read amaroK config file or no 'Collection Folders' value
         logger.error("setCollectionPath: cannot read amaroK config file "+str(detail))
         if not self.data.has_key("collfolders"):
            # there is no value in wabaroK config
            logger.critical("setCollectionPath: No Collection folder read from amarok nor wabarok, setting it to empty ()")
            self.data["collfolders"]=[]
         else:
            logger.warning("setCollectionPath: read from wabaroK config file (%s)",self.data["collfolders"].join(','))
   
   def setAmazonCoverPath(self,localprefix):
      """set path where to search amazon large covers"""
      logger.info("setAmazonCoverPath(%s)",localprefix)
      fn=os.path.join(localprefix,"share/apps/amarok/albumcovers")
      if os.path.isdir(fn):
         self.data['covers']=fn
         logger.debug('setAmazonCoverPath: set covers to %s',fn)
      else:
         self.data['covers']=""
         logger.error('setAmazonCoverPath: cannot find cover folder directory %s',fn)
      
   def savePickle(self):
      """save config data to file
      ** outdated **"""
      logger.error('savePickle: outdated function, we should never go in')
   #~ try:
      f=open(self['configfile'],"w")
      pickle.dump(self.data,f)
      f.close()
      #~ except:
         #~ print "Cannot save config"

   def saveText(self):
      logger.info("saveText()")
      waConf=ConfigParser.SafeConfigParser()
      waConf.add_section("General")
      for k in self.params.keys(): # for each possible parameter
         if self.data.has_key(k) and self.params[k][4]: # if exists && ToBeSaved
            if self.params[k][2]==ListType:
               waConf.set("General",k,','.join(self.data[k]))
            else: # str or int
               waConf.set("General",k,str(self.data[k]))
      try: # save to file
         waConf.write(open(self.data['configfile'],"w"))
      except Exception,detail:
         logger.error("saveText: cannot save text configuration file "+str(detail))
      

   def save(self):
      #self.savePickle()
      self.saveText()
      
   def __setitem__(self,key,item): self.data[key]=item
      
   def __getitem__(self,key): return self.data[key]

   def keys(self): return self.data.keys()

   def items(self): return self.data.items()  

   def values(self): return self.data.values()

   def __str__(self): return str(self.data)

   def guiKdialog(self,autoSave=1):
      """GUI to show/update object using kdialog
      autoSave: should the object be saved after gui closed ?"""
      logger.info("guiKdialog(%d)",autoSave)
      ended=0
      while not ended:
         menu = " ".join(["%s '%s (%s)'"%(k,v[0],str(self.data[k])) for k,v in self.params.items() if v[4]])
         file=os.popen("""%s --menu "Select parameter" %s \
                               --title "wabaroK Configuration" """ % (self.kdialog,menu) )
         retour=file.readlines()
         if len(retour)==0:
            ended=1
         else:
            key=retour[0].rstrip('\n')
            #print params[key]
            val=os.popen(self.params[key][1] % (self.kdialog,self.params[key][0],self.data[key])+" --title 'Change ...'").read().rstrip('\n')
            if val:
               if self.params[key][2] == IntType:
                  if val.isdigit():
                     self.data[key] = int(val)
                  else:
                     os.system("%s --error '%s must be only digits'"%(self.kdialog,self.params[key][0]))
               else:
                  # StringType
                  self.data[key]=val
      if autoSave: self.save()

if __name__=='__main__':
   logger.info("main")
   conf = wabaConf("test","test","test/wabarok.cfg")
   conf.read()
   logger.debug(conf)
   #conf.params['collfolders'][4]=1 # set to be saved
   #~ conf['imageWidth']="150"
   #~ conf['imageHeight']="150"
   #~ print conf
   conf.save()
   #conf.guiKdialog() # with autoSave
