#!/usr/bin/env python
# wabaroK : Browse amaroK collection via a web interface

import BaseHTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler
import SocketServer
import cgi,os,signal,sys,string
import amaroKdcop
from wabaConf import *
from inputFromAmarok import *
import logging,types
from base64 import urlsafe_b64encode,urlsafe_b64decode

# Handle URLs  :
# / or /index.rok /albums.rok => start page
# /album.rok?id=20 => display album of id 20
# /artists.rok => Show all artists
# /cover?url=path => display album cover of absolut path (b64, url safe)
# /mp3?url=path => stream mp3 file of absolut path (b64, url safe) | &kbps=16 => bitrate for low streaming
# /add2PL.xml?url=path => add path to amarok playlist (b64, url safe)
# /addA2PL.xml?id=20 => add album 20 to amarok playlist

# configuration objects
# Hey some global variables !!
# create logging object
logging.basicConfig(level=10, # 1st default logging level INFO
                    format='%(asctime)s %(name)s %(levelname)s %(message)s',
                    filename='/tmp/wabarok.log',
                    filemode='w')
logger=logging.getLogger('wabaroK   ')
# test if kde-config is executable
kde_config=os.popen('which kde-config').readline().rstrip()
if not os.access(kde_config,os.X_OK):
   logger.critical("Cannot execute kde-config (%s) will search only in ~/.kde/",kde_config)
else:
   logger.info("using %s",kde_config)
# test if lame is executable
lame=os.popen('which lame').readline().rstrip()
if not os.access(lame,os.X_OK):
   logger.critical("Cannot execute lame (%s) will not be able to reencode",lame)
else:
   logger.info("using %s",lame)
# create configuration object
localprefix=os.popen("%s --localprefix"%(kde_config)).read().rstrip('\n') or "~/.kde/"
conf = wabaConf(localprefix,
                os.path.expanduser(localprefix) + "share/apps/amarok/scripts/" + os.path.basename(sys.argv[0])[:-3]
                ,os.path.join(os.getcwd(),"wabaroK.cfg"))
conf.read()
logging.disable(int(conf['logging'])*10) # set logging level to config saved value
#logging.basicConfig(level=int(conf['logging'])*10)
#logger.setLevel(int(conf['logging'])*10)
#logger.info('Set logging level to %s',logging.getLevelName(int(conf['logging'])*10))
logger.warning("Configuration object: %s",conf)

class HTTPHandler (SimpleHTTPRequestHandler):
   global conf
   global logger
   server_version = "wabaroK-HTTP/0.2b3 20060513"
   dcop = amaroKdcop.amaroKdcop(conf['dbengine'])
   dico_stats=dcop.getStats()
   dico_artists = dcop.getArtists()
   dico_albums = dcop.getAlbums()
   # 1st char of names placed under # (like numbers) in listColumn
   listSpecChar = ('#','"','[','(','-','_','~','+')

   def sortbyvalue(self,dict):
      """ Return a list of (key, value) pairs, sorted by value.
      http://www.faqts.com/knowledge_base/view.phtml/aid/4364/fid/541
      Jul 5th, 2000 10:00
      Nathan Wallace, unknown unknown, Hans Nowak, Snippet 174, Python Snippet Support Team"""
      _swap2 = lambda (x,y): (y,x)
      _cmplower = lambda x, y: cmp(string.lower(x[0]), string.lower(y[0]))
      mdict = map(_swap2, dict.items())
      mdict.sort(_cmplower)
      mdict = map(_swap2, mdict)
      return mdict

   def translate_path(self,path):
      return conf['dataPath']+path

   def do_GET(self):
      if self.path.find('.rok') != -1 or self.path=="/":
         # This is a request that will return html
         if self.path.find('?') != -1:
            self.path, self.requete = self.path.split('?', 1)
         else:
            self.requete = ''
         self.reponseHTML()
      elif self.path.find(".xml") != -1:
         # action to be done on amaroK
         self.path, self.requete = self.path.split('?', 1)
         self.reponseXML()
      elif self.path[1:6] == "cover":
         # We will display a cover
         self.path, self.requete = self.path.split('?', 1)
         self.cover()
      elif self.path[1:4] == "mp3":
         # We will stream a mp3 file
         self.path, self.requete = self.path.split('?', 1)
         self.mp3()
      else:
         SimpleHTTPRequestHandler.do_GET(self)

   def do_POST(self):
      self.requete = self.rfile.read(int(self.headers['Content-Length']))
      self.reponseHTML()

   def reponseHTML(self):
      self.page = self.path[1:-4]
      self.args = dict(cgi.parse_qsl(self.requete))
      if self.args.has_key('url'):
         self.args['url']=urlsafe_b64decode(self.args['url'])
         logger.debug("reponseHTML: url decoded (self.args['url'])")
      self.send_response(200, 'OK')
      self.send_header('Content-type', 'text/html')
      self.end_headers()
      # send html page
      self.entete()
      self.menu()
      if self.page=="":
         self.page="index"
      if self.page in dir(self) and callable(getattr(self,self.page)):
            getattr(self,self.page)()
      self.piedepage()
   
   def reponseXML(self):
      # execute action on amaroK and return a xml file for result"""
      self.page = self.path[1:-4]
      self.args = dict(cgi.parse_qsl(self.requete))
      if self.args.has_key('url'):
         self.args['url']=urlsafe_b64decode(self.args['url'])
         logger.debug("reponseXML: url decoded (%s)"%(self.args['url']))
      self.send_response(200, 'OK')
      self.send_header('Content-type', 'text/xml')
      self.end_headers()
      # send xml result
      if self.page in dir(self) and callable(getattr(self,self.page)):
            getattr(self,self.page)()

   def index(self):
      """Show all albums"""
      self.albums()

   def album(self):
      """Show album content
      id=int"""
      # image = self.dcop.getImage(self.args['id'])
      album = self.dcop.getAlbumContent(self.args['id'],conf['covers'])
      logger.debug(album)
      self.wfile.write("""<h2>%(titre)s - %(auteur)s</h2>
                        <a href="#" title="Add album to amaorK Playlist" onclick=addA2PL("%(id)s")>PL</a><br/>"""
                        %{'titre':self.dico_albums[album['id']],
                          'auteur':self.dico_artists[album['artist']],
                          'id':self.args['id']})
      if album['cover'] != "":
         self.wfile.write("""<img src="/cover?url=%s" alt='Album Cover' align='left' width="%s" height="%s" hspace="10">\n"""%(urlsafe_b64encode(album['cover']),conf['imagewidth'],conf['imageheight']))
      for l in album['tracks']:
         try:
            (m,s)=divmod(int(l[2]),60)
            self.wfile.write("""%(num)s.<a title="Play song..." href="/mp3?url=%(url)s">%(titre)s</a> (%(min)02d:%(sec)02d)
                                <a href="/mp3?url=%(url)s&kbps=%(kbps)s" title="Play at low bitrate (%(kbps)s Kbps)">low</a> 
                                <a href="#" title="Add to amaroK Playlist" onclick=add2PL("%(url)s")>PL</a>
                                <br>\n"""%{'num':l[0].ljust(2),'url':urlsafe_b64encode(l[3]),'titre':l[1],'min':m,'sec':s,'kbps':conf['bitrate']})
         except IndexError :
            self.error("No songs in this album")

   def add2PL(self):
      """add a mp3 to amaroK Playlist
      result return as XmlHttpRequest"""
      url=self.args['url']
      logger.info("add2PL: %s",url)
      if not self.isInAuthorizedPath(url) or not os.path.isfile(url):
         logger.error('add2PL(%s) : not a file, or not authorized',url)
         result="%s not readable"%(url)
      else:
         ret = self.dcop.addToPlaylist(url)
         if ret:
            # error returned
            result = "error with %s : %s"%(url,ret)
         else:
            result="%s added to amaroK PL"%(url)
      self.wfile.write('<?xml version="1.0" standalone="yes"?><add2pl><result>%s</result></add2pl>' % (result))

   def addA2PL(self):
      """add a album to amaroK Playlist
      result return as XmlHttpRequest"""
      url=self.args['id']
      logger.info("addA2PL: %s",id)
      # get content
      album = self.dcop.getAlbumContent(self.args['id'],conf['covers'])
      result=0
      for l in album['tracks']:
         url=l[3]
         if not self.isInAuthorizedPath(url) or not os.path.isfile(url):
            logger.error('addA2PL(%s) : not a file, or not authorized',url)
         else:
            ret=self.dcop.addToPlaylist(url)
            if ret:
               # error returned
               logger.error('addA2PL(%s) : error during dcop call %s',url,ret)
            else:
               result+=1
      self.wfile.write("""<?xml version="1.0" standalone="yes"?><addA2pl><result>%s (%d tracks added to Playlist)</result></addA2pl>"""
                        % (album['album_name'],result))


   def albumsOld(self):
      """Show all albums old version"""
      #albums = self.dcop.getAlbums()
      self.wfile.write("<h2>Albums</h2>")
      list=[]
      header="""<div id="letters"><a href="1">#</a>"""
      letter="# "
      list+="""<h3><a name="1">#</a></h3>"""
      for l in self.sortbyvalue(self.dico_albums):
         logger.debug(l)
         if l[1] != "" and (l[1][0].lower() != letter and not l[1][0].isdigit() and not l[1][0] in self.listSpecChar):
            list+="""<h3><a name="%s">%s</a></h3>"""%(l[1][0].upper(),l[1][0].upper())
            header+="""<a href="#%s">%s</a> """%(l[1][0].upper(),l[1][0].upper())
            letter=l[1][0].lower()
         list+="""<a href="/album.rok?id=%s">%s</a><br>\n"""%(l[0],l[1])
      header+="</div>"
      list.insert(0,header)
      self.listColumn(list)

   def albums(self):
      """Show all albums"""
      self.wfile.write("<h2>Albums</h2>")
      getLetter=lambda l:l[1][0]
      getString=lambda l:"""<a href="/album.rok?id=%s">%s</a><br>\n"""%(l[0],l[1])
      existString=lambda l:l[1]!=""
      list = self.listAlphabetise(self.sortbyvalue(self.dico_albums),existString,getString,getLetter)
      self.listColumn(list,conf['column'])
       
   def artistsOld(self):
      """Show all artists"""
      self.wfile.write("<h2>Artists</h2>")
      #artists = self.dcop.getArtists()
      self.wfile.write("<div id='content_left'>")
      c=0
      for k in self.dico_artists.keys():
         if c == int(int(self.dico_stats['artists'])/2): self.wfile.write("</div><div id='content_right'>")
         self.wfile.write("""<a href="/artist.rok?id=%s">%s</a><br>\n"""%(k,self.dico_artists[k]))
         c+=1
      self.wfile.write("</div>")

   def artists(self):
      """Show all artists"""
      self.wfile.write("<h2>Artists</h2>")
      getLetter=lambda l:l[1][0]
      getString=lambda l:"""<a href="/artist.rok?id=%s">%s</a><br>\n"""%(l[0],l[1])
      existString=lambda l:l[1]!=""
      list = self.listAlphabetise(self.sortbyvalue(self.dico_artists),existString,getString,getLetter)
      self.listColumn(list,conf['column'])

   def songs(self):
      """Show all songs"""
      self.wfile.write("<h2>Songs</h2>")
      existString=lambda l:l[0]!=""
      getLetter=lambda l:l[0][0]
      getString=lambda l:"""<span class="bulle_song" title="%(auteur)s - %(album)s"><a href="/mp3?url=%(url)s">%(titre)s</a></span><br>\n"""%{'auteur':self.dico_artists[l[4]],'album':self.dico_albums[l[3]],'url':urlsafe_b64encode(l[2]),'titre':l[0]}
      songs = self.dcop.getSongs()
      songs.sort(lambda x, y: cmp(string.lower(x[0]), string.lower(y[0])))
      # ## songs must be sorted !! on songs[2] ## #
      list = self.listAlphabetise(songs,existString,getString,getLetter)
      self.listColumn(list,conf['column'])
      #~ self.wfile.write("<div id='content_left'>")
      #~ c=0
      #~ for l in songs:
         #~ if c == int(int(self.dico_stats['songs'])/2): self.wfile.write("</div><div id='content_right'>")
         #~ self.wfile.write("""<a href="/mp3?url=%s">%s</a><br>\n"""%(l[2],l[0]))
         #~ c+=1
      #~ self.wfile.write("</div>")

   def artist(self):
      """Show albums of an artist
      id=int"""
      self.wfile.write("<h2>Albums of %s</h2>"%(self.dico_artists[self.args['id']]))
      albums = self.dcop.getArtistAlbums(self.args['id'])
      for k in albums.keys():
         self.wfile.write("""<a href="/album.rok?id=%(id)s">%(name)s</a>
                             <a href="#" title="Add album to amaorK Playlist" onclick=addA2PL("%(id)s")>PL</a><br/>
                             <br>\n"""
                             %{'id':k,'name':albums[k]})

   def cover(self):
      """display a cover of an album"""
      self.args = dict(cgi.parse_qsl(self.requete))
      if self.args.has_key('url'):
         self.args['url']=urlsafe_b64decode(self.args['url'])
         logger.debug("cover: url decoded (self.args['url'])")
      if self.args['url'].lower().endswith('jpg'):
         content = 'image/jpg'
      else:
         content = 'image/png'
      self.displayFile(content,self.args['url'])

   def mp3(self):
      """stream a mp3 file"""
      self.args = dict(cgi.parse_qsl(self.requete))
      if self.args.has_key('url'):
         self.args['url']=urlsafe_b64decode(self.args['url'])
         logger.debug("mp3: url decoded (self.args['url'])")
      if self.args['url'].lower().endswith('ogg'):
         content = 'application/ogg'
      else:
         content = 'audio/mp3'
      if self.args.has_key('kbps') and int(self.args['kbps']) != 0:
         # we will reencode on the fly
         content = 'audio/mp3' # this will always be mp3
         logger.info("mp3(): %s %s (%s kbps)",content,self.args['url'],self.args['kbps'])
         self.displayFile(content,self.args['url'],int(self.args['kbps']))
      else:
         logger.info("mp3(): %s %s",content,self.args['url'])
         self.displayFile(content,self.args['url'])

   def displayFile(self,content,url,kbps=0):
      """read a file from disk and display or stream it"""
      if not self.isInAuthorizedPath(url) or not os.path.isfile(url):
         logger.error('displayFile(%s) : not a file, or not authorized',url)
         return
      self.send_response(200, 'OK')
      self.send_header('Content-type', content)
      self.end_headers()
      if kbps != 0:
         # reencode mp3
         f=os.popen("""%s --quiet -b %d "%s"  -"""%(lame,kbps,url))
      else:
         f = file(url,"r")
      self.copyfile(f, self.wfile)
      #if f:
         #self.wfile.write(f.read())
      logger.debug("stream ended (%s)",url)
      f.close()
   
   def listAlphabetise(self,data,existString,getString,getLetter):
      """Create a list alphabetized
      ie: with list of letters at the top and head letter at each new one
      data: ordered list of lines (each line is a list eg:(id,'name'))
      existString: callback: test existance of string to be displayed
      getString: callback: get a line, return string
      getLetter: callback: get a line, return 1st letter"""
      list=[]
      letter="#" # number and special characters
      header="""<div id="letters"><a href="1">%s</a> """%(letter)
      list.append("""<h3><a name="1">%s</a></h3>"""%(letter))
      for l in data:
         if existString(l):
            #print "%s,%s"%(l,getLetter(l))
            if (getLetter(l).lower() != letter and not getLetter(l).isdigit() and not getLetter(l) in self.listSpecChar):
               # new letter and not digit nor special char
               list.append("""<h3><a name="%s">%s</a></h3>"""%(getLetter(l).upper(),getLetter(l).upper()))
               header+="""<a href="#%s">%s</a> """%(getLetter(l).upper(),getLetter(l).upper())
               letter=getLetter(l).lower()
            list.append(getString(l))
      header+="</div><br/>"
      list.insert(0,header)
      return list
      
   def entete(self):
         """Display the head of the page"""
         self.wfile.write("""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
         <html>
         <head>
            <title>wabaroK Collection</title>
            <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
            <link rel="stylesheet" type="text/css" href="themes/%s.css">"""%(conf['theme']))
         # AjaX to add a file to amaroK playlist
         self.wfile.write("""<script language="javascript"  type="text/javascript">
         // XmlHTTPRequest //
         function basename (path) { return path.replace( /.*\//, "" ); }
         function handleHttpResponse()
         {
            if (http.readyState == 4)
            { /// Split the comma delimited response into an array/
               var xmlDocument = http.responseXML;
               var result = xmlDocument.getElementsByTagName('result').item(0).firstChild.data;
               //document.getElementById('result').value = result;
               // display something !
               alert(basename(result))
            }
         }
         function add2PL(mp3file) {
            http.open("GET", "add2PL.xml?url=" + mp3file,true);//+ escape(mp3file), true);
            http.onreadystatechange = handleHttpResponse;
            http.send(null);
         }
         function addA2PL(alb_id) {
            http.open("GET", "addA2PL.xml?id=" + alb_id,true);
            http.onreadystatechange = handleHttpResponse;
            http.send(null);
         }
         function getHTTPObject()
         {
            var xmlhttp;
            if (!xmlhttp && typeof XMLHttpRequest != 'undefined')
            {
               try { xmlhttp = new XMLHttpRequest(); }
               catch (e) { xmlhttp = false; }
            }
            return xmlhttp;
         }
         var http = getHTTPObject(); /// We create the HTTP Object/ 
         </script>""")
         self.wfile.write("""</head>
         <body><div id='header'><h1>wabaroK</h1></div>""")

   def stats(self):
      """display db statistics"""
      self.wfile.write("""<h2>Stats</h2>
      <ul class="menu_item">
         <li>Artists : %s</a></li>
         <li>Albums : %s</a></li>
         <li>Songs : %s</a></li>
      </ul>"""%(self.dico_stats['artists'],self.dico_stats['albums'],self.dico_stats['songs']))

   def menu(self):
      """Display the menu"""
      self.wfile.write("""<div id="nav"><h2>Menu</h2>
      <ul class="menu_item">
         <li><a href="/artists.rok">Artists</a></li>
         <li><a href="/albums.rok">Albums</a></li>
         <li><a href="/songs.rok">Songs</a></li>
      </ul>""")
      self.stats()
      self.wfile.write("""</div><div id="content">""")

   def piedepage(self):
      """Display the bottom of the page"""
      self.wfile.write("""</div><div id="footer">%s [<a href="http://kubuntu.free.fr/wabarok/">JJL</a>]</div></body></html>"""%(self.server_version))
   
   def listColumn(self,list,nbcol=2):
      """Display a list as nbcol column"""
      lcol = int(len(list)/nbcol) # col lenght
      self.wfile.write(list.pop(0)) # write list of letters
      self.wfile.write("\n<div id='content_col'>")
      c=0
      for l in list:
         if c >= lcol and l.find("<a name=")!=-1:
            # col length reached and head letter
            self.wfile.write("</div>\n<div id='content_col'>")
            c=0
         self.wfile.write(l)
         c+=1
      self.wfile.write("</div>")
      
   def error(self,message):
      """Display an error message"""
      self.wfile.write("""<h3>%s</h3>"""%(message))

   def isInAuthorizedPath(self,path):
      """Check if requested file (cover/mp3) is in
      Collection path : conf['collfolders']
      Amazon Cover path : conf['covers']"""
      paths=conf['collfolders']+[conf['covers'],]
      logger.debug("isInAuthorizedPath %s",paths)
      for p in paths:
         if path.startswith(p):
            logger.debug("isInAuthorizedPath(%s): %s",path,p)
            return 1
      logger.error("WARNING: try to access out of authorized path (%s)",path)
      return 0

def exit(signum,frame):
   global event
   logger.warning("exiting (%d) ...",signum)
   sys.stdin.close()
   sys.exit(signum)
   logger.shutdown()

class ThreadingHTTPServer (SocketServer.ThreadingMixIn,BaseHTTPServer.HTTPServer): pass

if __name__=='__main__':
   #~ sys.stdout=file("wabaout.log","w",0)
   #~ sys.stderr=sys.stdout

   signal.signal(signal.SIGTERM, exit)
   signal.signal(signal.SIGINT, exit)
   

   inputFromAmarok(sys.stdin,conf.guiKdialog).start()
   #conf.guiKdialog()

   httpd = ThreadingHTTPServer(('',conf['port'])   , HTTPHandler)
   httpd.serve_forever()

